from typing import BinaryIO, Iterator, Tuple

# Local import
from dissect.clfs.c_clfs import BlockHeader, c_clfs
from dissect.clfs.exceptions import InvalidRecordBlockError


class Container:
    """Main class for parsing the containers that belong to a BLF file parsed in an earlier stage.

    Args:
        fh: A file handle to a container file.
        offset: The offset to start parsing the container records.
    """

    def __init__(self, fh: BinaryIO, offset: int):
        self.fh = fh
        self.offset = offset

    def _open_block(self, offset: int) -> Tuple[BinaryIO, int]:
        """Open the blockheader of every block that is present within the given container.

        Returns:
            buf, cur_record_offset: Tuple containing a file-like object of the log block and the current offset.
        """

        try:
            log_block = BlockHeader(fh=self.fh, offset=offset)
        except EOFError:
            raise InvalidRecordBlockError("Invalid container block header, possibly corrupt/empty")

        cur_record_offset = log_block.header.RecordOffsets[0]

        buf = log_block.open()
        buf.seek(cur_record_offset)

        return buf, cur_record_offset

    def records(self) -> Iterator[Tuple[int, bytes, bytes]]:
        """Parse the records that are present within the log block."""

        log_block_offset = self.offset

        buf, cur_record_offset = self._open_block(log_block_offset)
        cur_record_header = c_clfs.RECORD_HEADER(buf)

        while True:
            # Data block
            if cur_record_header.Type & c_clfs.RecordType.ClfsDataRecord:
                cur_block_data = buf.read(cur_record_header.DataSize - cur_record_header.Offset)

            # Start of record header
            if cur_record_header.Type & c_clfs.RecordType.ClfsStartRecord:
                """
                This may seem odd to do, but the actual data offset is present in the start record for the record block.

                The record data is placed right after the header itself, meaning that if you read the entire block, you
                can subtract the offset (size of header) from the data size itself and be left with the record data.
                """

                # Advance to the next RECORD_HEADER
                next_record_header = c_clfs.RECORD_HEADER(buf)

                # The record data is present right after the record header, subtract the header size (offset field)
                # from the data size
                cur_record_data = buf.read(next_record_header.DataSize - next_record_header.Offset)

                yield log_block_offset + cur_record_offset, cur_record_data, cur_block_data

                # End of log sequence
                if next_record_header.LsnPrevious == 0:
                    break

                # Set the new block offset
                log_block_offset = next_record_header.LsnPrevious - 1
                cur_record_header = next_record_header
                continue

            # End of block, pointer to new block
            if cur_record_header.Type & c_clfs.RecordType.ClfsLastRecord:
                buf, cur_record_offset = self._open_block(log_block_offset)
                cur_record_header = c_clfs.RECORD_HEADER(buf)



import os

import pytest


def absolute_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def open_file(name, mode="rb"):
    with open(absolute_path(name), mode) as f:
        yield f


@pytest.fixture
def control_record_blf():
    yield from open_file("data/control_record.blf")


@pytest.fixture
def bad_control_record_blf():
    yield from open_file("data/bad_control_record.blf")


@pytest.fixture
def invalid_control_record_blf():
    yield from open_file("data/invalid_control_record.blf")


@pytest.fixture
def dummy_blf():
    yield from open_file("data/DRIVERS{53b39e70-18c4-11ea-a811-000d3aa4692b}.TM.blf")


@pytest.fixture
def dummy_container():
    yield from open_file(
        "data/DRIVERS{53b39e70-18c4-11ea-a811-000d3aa4692b}.TMContainer00000000000000000001.regtrans-ms"
    )



# Local imports
from dissect.clfs.c_clfs import BlockHeader, c_clfs


def test_record_header_c_definitions(control_record_blf):
    r_header = c_clfs.CLFS_LOG_BLOCK_HEADER(control_record_blf)

    assert r_header.MajorVersion == 0x15
    assert r_header.MinorVersion == 0x0
    assert r_header.Fixup == 0x1
    assert r_header.ClientId == 0x0
    assert r_header.TotalSectors == 0x2
    assert r_header.ValidSectors == 0x2
    assert r_header.Reserved1 == 0x0
    assert r_header.Checksum == 0xC64C824B
    assert r_header.Flags == 0x1
    assert r_header.Reserved2 == 0x0
    assert r_header.CurrentLsn.PhysicalOffset == 0xFFFFFFFF00000000
    assert r_header.NextLsn.PhysicalOffset == 0xFFFFFFFF00000000
    assert r_header.RecordOffsets[0] == 0x70
    assert r_header.FixupOffset == 0x3F8


def test_record_header(control_record_blf):
    logblock = BlockHeader(fh=control_record_blf, offset=0)

    assert logblock.header.MajorVersion == 0x15
    assert logblock.header.MinorVersion == 0x0
    assert logblock.header.Fixup == 0x1
    assert logblock.header.ClientId == 0x0
    assert logblock.header.TotalSectors == 0x2
    assert logblock.header.ValidSectors == 0x2
    assert logblock.header.Reserved1 == 0x0
    assert logblock.header.Checksum == 0xC64C824B
    assert logblock.header.Flags == 0x1
    assert logblock.header.Reserved2 == 0x0
    assert logblock.header.CurrentLsn.PhysicalOffset == 0xFFFFFFFF00000000
    assert logblock.header.NextLsn.PhysicalOffset == 0xFFFFFFFF00000000
    assert logblock.header.RecordOffsets[0] == 0x70
    assert logblock.header.FixupOffset == 0x3F8



import pytest

# Local imports
from dissect.clfs.blf import BLF
from dissect.clfs.exceptions import InvalidBLFError


def test_validate(control_record_blf):
    BLF(fh=control_record_blf)


def test_validate_fail(invalid_control_record_blf):
    with pytest.raises(InvalidBLFError):
        BLF(fh=invalid_control_record_blf)



import io
from typing import BinaryIO

# External dependencies
from dissect.cstruct import cstruct

clfs_def = """
/* ======== Generic Windows ======== */
flag FILE_ATTRIBUTES : USHORT {
    READONLY                    = 0x00000001,
    HIDDEN                      = 0x00000002,
    SYSTEM                      = 0x00000004,
    DIRECTORY                   = 0x00000010,
    ARCHIVE                     = 0x00000020,
    DEVICE                      = 0x00000040,
    NORMAL                      = 0x00000080,
    TEMPORARY                   = 0x00000100,
    SPARSE_FILE                 = 0x00000200,
    REPARSE_POINT               = 0x00000400,
    COMPRESSED                  = 0x00000800,
    OFFLINE                     = 0x00001000,
    NOT_CONTENT_INDEXED         = 0x00002000,
    ENCRYPTED                   = 0x00004000,
    VIRTUAL                     = 0x00010000,
    INVALID_FILE_ATTRIBUTES     = 0xFFFFFFFF,
};

typedef UCHAR CLFS_CLIENT_ID;

typedef DWORD CLFS_RECORD_INDEX;
typedef DWORD CLFS_CONTAINER_ID;

// Taken from clfslsn.h (Windows SDK)
typedef union CLFS_LSN {
    struct {
        CLFS_RECORD_INDEX   RecordIndex;
        CLFS_CONTAINER_ID   ContainerId;
    } Offset;

    ULONGLONG PhysicalOffset;
};

typedef enum CLFS_NODE_TYPE : DWORD {
    FCB                      = 0xC1FDF001,
    VCB                      = 0xC1FDF002,
    CCB                      = 0xC1FDF003,
    REQ                      = 0xC1FDF004,
    CCA                      = 0xC1FDF005,
    SYMBOL                   = 0xC1FDF006,
    CLIENT_CONTEXT           = 0xC1FDF007,
    CONTAINER_CONTEXT        = 0xC1FDF008,
    SHARED_SECURITY_CONTEXT  = 0xC1FDF00D,
    DEVICE_EXTENSION         = 0xC1FDF009,
    MARSHALING_AREA          = 0xC1FDF00A,
    ARCHIVE_CONTEXT          = 0xC1FDF00C,
    SCAN_CONTEXT             = 0xC1FDF00E,
    LOG_READ_IOCB            = 0xC1FDF00F,
    LOG_WRITE_IOCB           = 0xC1FDF010,
};

typedef struct CLFS_NODE_ID {
    CLFS_NODE_TYPE Type;
    DWORD Node;
};

/* ======== Log Block Header ======== */

typedef enum CLFS_LOG_BLOCK_FLAGS : DWORD {
    RESET             = 0x00000000,
    ENCODED           = 0x00000001,
    DECODED           = 0x00000002,
    LATCHED           = 0x00000004,
    TRUNCATE_DISCARD  = 0x00000008,
};

typedef struct CLFS_LOG_BLOCK_HEADER {
    UCHAR MajorVersion;
    UCHAR MinorVersion;
    UCHAR Fixup;
    UCHAR ClientId;
    USHORT TotalSectors;
    USHORT ValidSectors;
    DWORD Reserved1;
    DWORD Checksum;
    CLFS_LOG_BLOCK_FLAGS Flags;
    DWORD Reserved2;
    CLFS_LSN CurrentLsn;
    CLFS_LSN NextLsn;
    DWORD RecordOffsets[16];
    DWORD FixupOffset;
};

/* ======== Control Record ======== */

typedef struct CLFS_METADATA_RECORD_HEADER {
    ULONGLONG DumpCount;
};

typedef enum CLFS_EXTEND_STATE : DWORD {
    ClfsExtendStateNone             = 0x00000000,
    ClfsExtendStateExtendingFsd     = 0x00000001,
    ClfsExtendStateFlushingBlock    = 0x00000002,
};

typedef enum CLFS_TRUNCATE_STATE : DWORD {
    ClfsTruncateStateNone                   = 0x00000000,
    ClfsTruncateStateModifyingStream        = 0x00000001,
    ClfsTruncateStateSavingOwner            = 0x00000002,
    ClfsTruncateStateModifyingOwner         = 0x00000003,
    ClfsTruncateStateSavingDiscardBlock     = 0x00000004,
    ClfsTruncateStateModifyingDiscardBlock  = 0x00000005,
};

typedef struct CLFS_TRUNCATE_CONTEXT {
    CLFS_TRUNCATE_STATE TruncateState;
    UCHAR Clients;
    UCHAR Client;
    USHORT TruncateField;
    CLFS_LSN LsnOwnerPage;
    CLFS_LSN LsnLastOwnerPage;
    ULONGLONG InvalidSector;
};

typedef enum CLFS_METADATA_BLOCK_TYPE : DWORD {
    ClfsMetaBlockControl          = 0x00000000,
    ClfsMetaBlockControlShadow    = 0x00000001,
    ClfsMetaBlockGeneral          = 0x00000002,
    ClfsMetaBlockGeneralShadow    = 0x00000003,
    ClfsMetaBlockScratch          = 0x00000004,
    ClfsMetaBlockScratchShadow    = 0x00000005,
};

typedef struct CLFS_METADATA_BLOCK {
    union {
        ULONGLONG Image;
        ULONGLONG Alignment;
    };

    DWORD Image;
    DWORD Offset;
    CLFS_METADATA_BLOCK_TYPE Type;
    DWORD Padding;
};

typedef struct CLFS_CONTROL_RECORD {
    CLFS_METADATA_RECORD_HEADER RecordHeader;
    ULONGLONG Magic;
    UCHAR Version;
    UCHAR Reserved1;
    UCHAR Reserved2;
    UCHAR Reserved3;
    CLFS_EXTEND_STATE ExtendState;
    USHORT ExtendBlock;
    USHORT FlushBlock;
    DWORD NewBlockSectors;
    DWORD ExtendStartSectors;
    DWORD ExtendSectors;
    CLFS_TRUNCATE_CONTEXT Truncate;
    DWORD Blocks;
    DWORD Reserved4 ;
    CLFS_METADATA_BLOCK RgBlocks[Blocks];
};

/* ======== Base Record ======== */

typedef struct CLFS_BASE_RECORD_HEADER {
    CLFS_METADATA_RECORD_HEADER RecordHeader;
    UCHAR IdLog[16];
    ULONGLONG ClientSymbolTable[11];
    ULONGLONG ContainerSymbolTable[11];
    ULONGLONG SecuritySymbolTable[11];
    DWORD NextContainer;
    DWORD NextClient;
    DWORD FreeContainers;
    DWORD ActiveContainers;
    DWORD FreeContainersCount;
    DWORD BusyContainers;
    DWORD ClientContainers[124];
    DWORD ContainerArray[1024];
    DWORD SymbolZone;
    DWORD Sector;
    UCHAR Unused1;
    UCHAR Unused2;
    UCHAR Unused3;
    UCHAR Usn;
    UCHAR Clients;
};

typedef struct CLFS_HASH_SYM {
    CLFS_NODE_ID NodeId;
    DWORD UlHash;
    DWORD CbHash;
    ULONGLONG Below;
    ULONGLONG Above;
    DWORD SymbolName;
    USHORT Offset;
    USHORT Deleted;
};

typedef enum CLFS_LOG_STATE : DWORD {
    UNINITIALIZED    = 0x00000001,
    INITIALIZED      = 0x00000002,
    ACTIVE           = 0x00000004,
    PENDING_DELETE   = 0x00000008,
    PENDING_ARCHIVE  = 0x00000010,
    SHUTDOWN         = 0x00000020,
    MULTIPLEXED      = 0x00000040,
    SECURE           = 0x00000080,
};

typedef struct CLFS_CLIENT_CONTEXT {
    CLFS_NODE_ID NodeId;
    CLFS_CLIENT_ID ClientId;
    UCHAR Unknown1;
    FILE_ATTRIBUTES FileAttributes;
    DWORD FlushThreshold;
    ULONGLONG Unknown2[5]; // could be timestamp?
    CLFS_LSN LsnArchiveTail;
    CLFS_LSN LsnBase;
    CLFS_LSN LsnFlush;
    CLFS_LSN LsnLast;
    CLFS_LSN LsnPhysicalBase;
    CLFS_LSN LsnUnused1;
    CLFS_LSN LsnUnused2;
    CLFS_LOG_STATE State;
    ULONGLONG SecurityContext;
};

typedef struct CLFS_CONTAINER_CONTEXT {
    CLFS_NODE_ID NodeId;
    ULONGLONG Container;
    CLFS_CONTAINER_ID ContainerId;
    CLFS_CONTAINER_ID QueueId;
    ULONGLONG Alignment;
    UCHAR CurrentUsn;
    UCHAR State;
    DWORD PreviousOffset;
    DWORD NextOffset;
};

// This should only ever be present in-memory
typedef struct CLFS_SHARED_SECURITY_CONTEXT {
    CLFS_NODE_ID NodeId;
    DWORD Ref;
    DWORD RefActive;
    DWORD DescriptorOffset;
    DWORD Descriptor;
    UCHAR RgSecurityDescriptor[0]; // initialize as 0 for now, we're only parsing persistent files
};

/* ======== Truncate Record ======== */

typedef struct CLFS_TRUNCATE_RECORD_HEADER {
    CLFS_METADATA_RECORD_HEADER RecordHeader;
    DWORD ClientChangeOffset;
    DWORD OwnerPageOffset;
};

typedef struct CLFS_SECTOR_CHANGE {
    DWORD InitializedSector;
    DWORD Unused;
    UCHAR Sector[512];    // CLFS sector size
};

typedef struct CLFS_TRUNCATE_CLIENT_CHANGE {
    CLFS_CLIENT_ID ClientId;
    CLFS_LSN Lsn;
    CLFS_LSN LsnClient;
    CLFS_LSN LsnRestart;
    USHORT Length;
    USHORT OldLength;
    DWORD Sectors;
    CLFS_SECTOR_CHANGE RgSectors[Sectors];
};

/* ======== Container Record Header ======== */

flag RecordType : DWORD {
    ClfsNullRecord          = 0x00000000, // Null record
    ClfsDataRecord          = 0x00000001, // The log record contains client data
    ClfsRestartRecord       = 0x00000002, // The log record is a restart record
    ClfsStartRecord         = 0x00000004, // Start of continuation record
    ClfsEndRecord           = 0x00000008, // End of continuation record
    ClfsContinuationRecord  = 0x00000010, // Continuation record
    ClfsLastRecord          = 0x00000020, // The last record in the log block
};

typedef struct RECORD_HEADER {
    ULONGLONG LsnVirtual;
    ULONGLONG LsnUndoNext;
    ULONGLONG LsnPrevious;
    DWORD DataSize;
    DWORD Unknown;
    USHORT RecordFlags;
    USHORT Offset;
    RecordType Type;
};
"""

c_clfs = cstruct()
c_clfs.load(clfs_def)

SECTOR_SIZE = 512
CLFS_CONTROL_RECORD_MAGIC_VALUE = 0xC1F5C1F500005F1C


class BlockHeader:
    """Main class to parse the block headers.

    Args:
        fh: A file-like object.
        offset: Offset to start reading the block header from.
    """

    def __init__(self, fh: BinaryIO, offset: int):
        self.offset = offset

        fh.seek(self.offset)
        self.header = c_clfs.CLFS_LOG_BLOCK_HEADER(fh)

        fh.seek(self.offset)
        data = bytearray(fh.read(self.header.TotalSectors * SECTOR_SIZE))
        view = memoryview(data)

        fixup = view[self.header.FixupOffset :]
        ptr = view[SECTOR_SIZE - 2 :]

        for _ in range(self.header.TotalSectors):
            ptr[:2] = fixup[:2]
            fixup = fixup[2:]
            ptr = ptr[SECTOR_SIZE:]

        self.data = bytes(data)

    def open(self):
        """Return a file-like object of the block header."""

        return io.BytesIO(self.data)



from collections import namedtuple

# Local imports
from dissect.clfs.container import Container

Data = namedtuple("Data", ["offset", "r_data", "b_data"])


def test_container_records(dummy_container):
    records = []

    trans = Container(fh=dummy_container, offset=36864)

    for r_offset, r_data, b_data in trans.records():
        records.append(Data(offset=r_offset, r_data=r_data, b_data=b_data))

    assert len(records) == 12

    assert records[0].offset == 36976

    expected_first_record_data = bytes.fromhex("000000000000000004010000762f16000519ea11a810000d3aa41ef300000000")
    assert records[0].r_data == expected_first_record_data

    expected_first_block_data = bytes.fromhex(
        """
        04 01 00 00 04 00 00 00
        be 8e 6b c1 61 db ec 11
        a4 ba 00 50 56 ef c5 14
        0d 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 80 02 01 00 00 10
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        """
    )
    assert records[0].b_data == expected_first_block_data

    expected_last_record_data = bytes.fromhex("000000000000000004010000762f16000519ea11a810000d3aa41ef300000000")
    assert records[-1].r_data == expected_last_record_data

    expected_last_block_data = bytes.fromhex(
        """
        04 01 00 00 04 00 00 00
        e1 b9 29 a6 ea d7 eb 11
        a4 6b 3c 22 fb 13 6b f1
        21 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 80 02 01 00 00 10
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        """
    )
    assert records[-1].b_data == expected_last_block_data



from collections import namedtuple
from typing import BinaryIO, Iterator

from dissect.clfs.c_clfs import CLFS_CONTROL_RECORD_MAGIC_VALUE, BlockHeader, c_clfs
from dissect.clfs.exceptions import (
    InvalidBLFError,
    InvalidContextError,
    InvalidRecordBlockError,
)

Context = namedtuple("Context", ["symbol_table", "type"])
Container = namedtuple("Container", ["name", "size", "id", "type"])
Stream = namedtuple(
    "Stream",
    [
        "name",
        "id",
        "file_attributes",
        "type",
        "lsn_archive_tail",
        "lsn_base",
        "lsn_last",
        "lsn_flush",
        "lsn_physical_base",
        "offset",
    ],
)


class ControlRecord:
    """Read the Control Record of the BLF.

    The control record is used to specify where the other blocks that make up the BLF are being stored within the
    given file. Each entry will have their image size, as well as the offset which is relative to the beginning of
    the file. This header also contains the CLFS magic value which in turn can be used to validate the given file
    (0xC1F5C1F500005F1C).

    There are 3 different types of blocks:

    - Control Block
    - General Block
    - Metadata Block

    Each of the 3 different blocks has another shadow block accompanied that is storing the information from the
    previous transaction on this block. A shadow block can be identified by the odd number in the block_type field.

    Args:
        fh: A file-like object to a BLF file.
        offset: Offset to start reading the control records.
    """

    def __init__(self, fh: BinaryIO, offset: int):
        try:
            self.logblock = BlockHeader(fh=fh, offset=offset)
        except (EOFError, AttributeError):
            raise InvalidRecordBlockError("Invalid control record block header, possibly corrupt/empty")

        record_offest = self.logblock.header.RecordOffsets[0]
        logblock_fh = self.logblock.open()
        logblock_fh.seek(record_offest)

        try:
            self.record = c_clfs.CLFS_CONTROL_RECORD(logblock_fh)
        except (EOFError, AttributeError):
            raise InvalidRecordBlockError("Invalid control record, possibly corrupt/empty")

    @property
    def valid(self) -> bool:
        """Return if the control record magic is valid."""
        return self.record.Magic == CLFS_CONTROL_RECORD_MAGIC_VALUE


class BaseRecord:
    """Function to parse the base record block.

    The base record block contains information about the different containers and which clients (streams) are using
    these. The first entries in the client and container symbol tables are the ones used, the full array can have
    more than just those values, but at the time of writing it is unclear if these are artefacts from earlier
    transactions in the log.

    For every context (client, container, security) the symbol table is parsed to extract the stream and container
    data. At this point we don't do anything with the security context yet as this seems only used for in-memory
    structures of CLFS.

    Args:
        fh: A file-like object to a BLF file.
        offset: Offset to start reading the base records.
        block_type: Type of CLFS block to parse.
    """

    def __init__(self, fh: BinaryIO, offset: int, block_type: int):
        self.block_type = block_type

        self.containers = []
        self.streams = []

        try:
            self.logblock = BlockHeader(fh=fh, offset=offset)
        except EOFError:
            raise InvalidRecordBlockError("Invalid base record block header, possibly corrupt/empty")

        record_offset = self.logblock.header.RecordOffsets[0]
        logblock_fh = self.logblock.open()
        logblock_fh.seek(record_offset)

        self.record = c_clfs.CLFS_BASE_RECORD_HEADER(logblock_fh)

        # Create the 3 contexts as a named tuple for more descriptive parsing
        contexts = [
            Context(
                symbol_table=self.record.ClientSymbolTable,
                type=c_clfs.CLFS_NODE_TYPE.CLIENT_CONTEXT,
            ),
            Context(
                symbol_table=self.record.ContainerSymbolTable,
                type=c_clfs.CLFS_NODE_TYPE.CONTAINER_CONTEXT,
            ),
            Context(
                symbol_table=self.record.SecuritySymbolTable,
                type=c_clfs.CLFS_NODE_TYPE.SHARED_SECURITY_CONTEXT,
            ),
        ]

        for ctx in contexts:
            self._symbol_table(
                sym_table=ctx.symbol_table, ctx_type=ctx.type, logblock_fh=logblock_fh, offset=record_offset
            )

    def _symbol_table(
        self, sym_table: list, ctx_type: c_clfs.CLFS_NODE_TYPE, logblock_fh: BinaryIO, offset: int
    ) -> None:
        """Function to parse the symbol tables.

        Clients, containers, and shared security contexts in the Base Log File are represented by symbols, which are
        preceded by the ClfsHashSym structure.

        The key takeaways from this structure are the symbol name which is the actualy container name the context is
        related to. This is also the name of the file on disk that is being used to write the log transaction to.

        For every stored context there is an offset noted in this structure that should be parsed seperately.

        Args:
            sym_table: The offset as denoted in the symbol table.
            ctx_type: Description of the type of context (CLIENT_CONTEXT, CONTAINER_CONTEXT, SHARED_SECURITY_CONTEXT).
            logblock_fh: A file-like object of the log block in which the symbol table resides.
            offset: Offset with start of the record.
        """
        for sym_tbl_offset in sym_table:
            if sym_tbl_offset == 0:
                continue

            # Seek towards the start of the symbol table
            logblock_fh.seek(offset + sym_tbl_offset)

            sym_tbl = c_clfs.CLFS_HASH_SYM(logblock_fh)

            if sym_tbl.NodeId.Type != c_clfs.CLFS_NODE_TYPE.SYMBOL:
                raise InvalidContextError(f"Invalid NodeId type: {sym_tbl.NodeId.Type}")

            # It looks like the size of the symbol names is not stored anywhere and
            # simply read until a null-terminator is found...
            logblock_fh.seek(offset + sym_tbl.SymbolName)
            symbol_name = b"".join(iter(lambda: logblock_fh.read(2), b"\x00\x00")).decode("utf-16")

            ctx_offset = offset + sym_tbl.Offset

            # Parse the context based on the context type
            if ctx_type == c_clfs.CLFS_NODE_TYPE.CLIENT_CONTEXT:
                self.streams.append(
                    self._client_context(logblock_fh=logblock_fh, client_ctx_offset=ctx_offset, name=symbol_name)
                )

            elif ctx_type == c_clfs.CLFS_NODE_TYPE.CONTAINER_CONTEXT:
                self.containers.append(
                    self._container_context(logblock_fh=logblock_fh, container_ctx_offset=ctx_offset, name=symbol_name)
                )

            elif ctx_type == c_clfs.CLFS_NODE_TYPE.SHARED_SECURITY_CONTEXT:
                self._security_context(logblock_fh=logblock_fh, security_ctx_offset=ctx_offset)
            else:
                raise InvalidContextError(f"Invalid context type: {ctx_type}")

    def _client_context(self, logblock_fh: BinaryIO, client_ctx_offset: int, name: str) -> None:
        """Function to parse client context structures.

        The client context relates to the stream that is being used for the current transaction. The client_id
        field is used to match the container that the stream of records should be written to.

        The LSN (Logical Sequence Number) base holds the offset that is relative of the start of the container
        and points to the last transaction (actually it points to the one that is currently being transacted,
        but I'm not 100% sure about this) and is used as a starting offset when reading the container.

        NOTE: a lot more of the client context is currently being stored in the streams list, this is mostly
        because I think this might actually be valuable information for the complete dissect implementation.

        Args:
            logblock_fh: A file-like object of the log block in which the client context resides.
            client_ctx_offset: Starting offset of the client context.
            name: Filename to which the stream for the current transaction belongs to.

        Returns:
            Stream: A named tuple containing the necessary information about the client context.
        """
        logblock_fh.seek(client_ctx_offset)
        client_ctx_record = c_clfs.CLFS_CLIENT_CONTEXT(logblock_fh)

        if client_ctx_record.NodeId.Type != c_clfs.CLFS_NODE_TYPE.CLIENT_CONTEXT:
            raise InvalidContextError(f"Invalid NodeId type: {client_ctx_record.NodeId.Type}")

        return Stream(
            name=name,
            id=client_ctx_record.ClientId,
            file_attributes=client_ctx_record.FileAttributes,
            type=self.block_type,
            lsn_archive_tail=client_ctx_record.LsnArchiveTail,
            lsn_base=client_ctx_record.LsnBase,
            lsn_last=client_ctx_record.LsnLast,
            lsn_flush=client_ctx_record.LsnFlush,
            lsn_physical_base=client_ctx_record.LsnPhysicalBase,
            offset=client_ctx_record.LsnPhysicalBase.Offset.RecordIndex - 1,
        )

    def _container_context(self, logblock_fh: BinaryIO, container_ctx_offset: int, name: str) -> None:
        """Function to parse container context structures.

        The container context is important to keep track of as the container ID should be in line
        with the stream ID and is thus needed to make sure that the offsets we're using are indeed
        meant for the right container if we want to parse the data correctly.

        Args:
            logblock_fh: A file-like object of the log block in which the container context resides.
            container_ctx_offset: Starting offset of the container context.
            name: Filename to which the stream for the current transaction belongs to.

        Returns:
            Container: A namedtuple containing the necessary information about the container context.
        """
        logblock_fh.seek(container_ctx_offset)
        container_ctx_record = c_clfs.CLFS_CONTAINER_CONTEXT(logblock_fh)

        if container_ctx_record.NodeId.Type != c_clfs.CLFS_NODE_TYPE.CONTAINER_CONTEXT:
            raise InvalidContextError(f"Invalid NodeId type: {container_ctx_record.NodeId.Type}")

        return Container(
            name=name,
            size=container_ctx_record.Container,
            id=container_ctx_record.ContainerId,
            type=self.block_type,
        )

    def _security_context(self, logblock_fh: BinaryIO, security_ctx_offset: int) -> None:
        """Function to parse the security context structures.

        NOTE: this is an in-memory structure and will be parsed at a later stage (dissect.mem bby)

        Args:
            logblock_fh: A file-like object of the log block in which the security context resides.
            security_ctx_offset: Starting offset of the security context.
        """
        logblock_fh.seek(security_ctx_offset)
        security_ctx_record = c_clfs.CLFS_SHARED_SECURITY_CONTEXT(logblock_fh)

        if security_ctx_record.NodeId.Type != c_clfs.CLFS_NODE_TYPE.SHARED_SECURITY_CONTEXT:
            raise InvalidContextError(f"Invalid NodeId type: {security_ctx_record.NodeId.Type}")


class TruncateRecord:
    """Parser for the truncate records within a BLF if they exist.

    Args:
        fh: A file-like object to a BLF file.
        offset: Offset to start reading the truncate records.
    """

    def __init__(self, fh: BinaryIO, offset: int):
        try:
            self.logblock = BlockHeader(fh=fh, offset=offset)
        except EOFError:
            raise InvalidRecordBlockError("Invalid truncate record block header, possibly corrupt/empty")

        record_offset = self.logblock.header.RecordOffsets[0]
        logblock_fh = self.logblock.open()
        logblock_fh.seek(record_offset)

        self.record = c_clfs.CLFS_TRUNCATE_RECORD_HEADER(logblock_fh)

        if self.record.ClientChangeOffset == 0:
            # CLFS_TRUNCATE_RECORD_HEADER = 16 bytes
            record_offset += 0x10
        else:
            record_offset += self.record.ClientChangeOffset


class BLF:
    """Main class of dissect.clfs. Parsing of BLF and information regarding the associated containers starts here.

    Args:
        fh: A file-like object to a BLF file.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        self.c_record = ControlRecord(fh=self.fh, offset=0)

        if not self.c_record.valid:
            raise InvalidBLFError("Invalid BLF file, possibly corrupt/empty")

        self.metablocks = self.c_record.record.RgBlocks

    def control_records(self) -> Iterator[ControlRecord]:
        """Yield the associated control records."""
        for metablock in self.metablocks:
            if metablock.Type in (
                c_clfs.CLFS_METADATA_BLOCK_TYPE.ClfsMetaBlockControl,
                c_clfs.CLFS_METADATA_BLOCK_TYPE.ClfsMetaBlockControlShadow,
            ):
                yield ControlRecord(fh=self.fh, offset=metablock.Offset)

    def base_records(self) -> Iterator[BaseRecord]:
        """Yield the associated base records.

        The base records hold most of the information regarding the parsing of the associated containers.
        """
        for metablock in self.metablocks:
            if metablock.Type in (
                c_clfs.CLFS_METADATA_BLOCK_TYPE.ClfsMetaBlockGeneral,
                c_clfs.CLFS_METADATA_BLOCK_TYPE.ClfsMetaBlockGeneralShadow,
            ):
                yield BaseRecord(fh=self.fh, offset=metablock.Offset, block_type=metablock.Type)

    def truncate_records(self) -> Iterator[TruncateRecord]:
        """Yield the truncate records.

        This has not been encountered yet.
        """
        for metablock in self.metablocks:
            if metablock.Type in (
                c_clfs.CLFS_METADATA_BLOCK_TYPE.ClfsMetaBlockScratch,
                c_clfs.CLFS_METADATA_BLOCK_TYPE.ClfsMetaBlockScratchShadow,
            ):
                yield TruncateRecord(fh=self.fh, offset=metablock.Offset)



class Error(Exception):
    pass


class InvalidRecordBlockError(Error):
    """Exception raised when the record block contains invalid data"""


class InvalidBLFError(Error):
    """Exception raised if the validation of the BLF file fails"""


class InvalidSymbolTableError(Error):
    """Exception raised when the symbol table(s) fail to parse"""


class InvalidContextError(Error):
    """Exception raised when the context type doesn't match the context to be parsed"""



import pytest

from dissect.clfs.blf import BLF

# Local imports
from dissect.clfs.c_clfs import c_clfs
from dissect.clfs.exceptions import InvalidRecordBlockError


def test_control_record_c_definitions(control_record_blf):
    # Seek to start of the control record
    control_record_blf.seek(0x70)

    c_record = c_clfs.CLFS_CONTROL_RECORD(control_record_blf)

    assert c_record.RecordHeader.DumpCount == 0x1
    assert c_record.Magic == 0xC1F5C1F500005F1C
    assert c_record.Version == 0x1
    assert c_record.Reserved1 == 0x0
    assert c_record.Reserved2 == 0x0
    assert c_record.Reserved3 == 0x0
    assert c_record.ExtendState == 0x0
    assert c_record.ExtendBlock == 0x0
    assert c_record.FlushBlock == 0x0
    assert c_record.NewBlockSectors == 0x0
    assert c_record.ExtendStartSectors == 0x0
    assert c_record.ExtendSectors == 0x0
    assert c_record.Blocks == 0x6


def test_control_record_blf(control_record_blf):
    blf = BLF(fh=control_record_blf)

    assert blf.c_record.record.RecordHeader.DumpCount == 0x1
    assert blf.c_record.record.Magic == 0xC1F5C1F500005F1C
    assert blf.c_record.record.Version == 0x1
    assert blf.c_record.record.Reserved1 == 0x0
    assert blf.c_record.record.Reserved2 == 0x0
    assert blf.c_record.record.Reserved3 == 0x0
    assert blf.c_record.record.ExtendState == 0x0
    assert blf.c_record.record.ExtendBlock == 0x0
    assert blf.c_record.record.FlushBlock == 0x0
    assert blf.c_record.record.NewBlockSectors == 0x0
    assert blf.c_record.record.ExtendStartSectors == 0x0
    assert blf.c_record.record.ExtendSectors == 0x0
    assert blf.c_record.record.Blocks == 0x6


def test_control_record_fail(bad_control_record_blf):
    with pytest.raises(InvalidRecordBlockError):
        BLF(fh=bad_control_record_blf)
